import json
from urllib.parse import urljoin

import anyio
import httpx
import pytest

from mcp_stata.server import get_ui_channel, run_command


pytestmark = [pytest.mark.requires_stata, pytest.mark.integration]


def _run_command_sync(code: str) -> str:
    async def _main() -> str:
        return await run_command(code)

    return anyio.run(_main)


def test_ui_http_auth_and_basic_endpoints():
    _run_command_sync("sysuse auto, clear")

    info = json.loads(get_ui_channel())
    base = info["baseUrl"]
    token = info["token"]

    # 401 when missing token
    r = httpx.get(urljoin(base, "/v1/dataset"))
    assert r.status_code == 401

    headers = {"Authorization": f"Bearer {token}"}

    r2 = httpx.get(urljoin(base, "/v1/dataset"), headers=headers)
    assert r2.status_code == 200
    payload = r2.json()
    assert payload["dataset"]["n"] > 0
    assert payload["dataset"]["k"] > 0

    r3 = httpx.get(urljoin(base, "/v1/vars"), headers=headers)
    assert r3.status_code == 200
    vars_payload = r3.json()
    names = {v["name"] for v in vars_payload["variables"]}
    assert "price" in names


def test_ui_http_paging_and_views_filtering():
    _run_command_sync("sysuse auto, clear")

    info = json.loads(get_ui_channel())
    base = info["baseUrl"]
    token = info["token"]
    headers = {"Authorization": f"Bearer {token}"}

    ds = httpx.get(urljoin(base, "/v1/dataset"), headers=headers).json()["dataset"]

    page_req = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "limit": 20,
        "vars": ["price", "mpg", "make"],
        "includeObsNo": True,
        "maxChars": 200,
    }

    page = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=page_req).json()
    assert page["view"]["returned"] == 20
    assert page["vars"][0] == "_n"
    assert len(page["rows"]) == 20

    view_req = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "filterExpr": "foreign == 1",
    }
    view = httpx.post(urljoin(base, "/v1/views"), headers=headers, json=view_req)
    assert view.status_code == 200
    view_payload = view.json()
    view_id = view_payload["view"]["id"]
    assert view_payload["view"]["filteredN"] < ds["n"]

    view_page_req = {
        "offset": 0,
        "limit": 10,
        "vars": ["price", "foreign"],
        "includeObsNo": True,
        "maxChars": 200,
    }
    vpage = httpx.post(urljoin(base, f"/v1/views/{view_id}/page"), headers=headers, json=view_page_req)
    assert vpage.status_code == 200
    vpayload = vpage.json()
    assert vpayload["view"].get("viewId") == view_id
    assert vpayload["view"].get("filteredN") is not None

    # Ensure filter applied: foreign column should be 1 or "1" depending on storage.
    foreign_idx = vpayload["vars"].index("foreign")
    for row in vpayload["rows"]:
        val = row[foreign_idx]
        assert str(val) in {"1", "1.0"}

    # Cleanup
    d = httpx.delete(urljoin(base, f"/v1/views/{view_id}"), headers=headers)
    assert d.status_code == 200


def test_ui_http_page_limit_validation():
    """Test that limit parameter is properly validated"""
    _run_command_sync("sysuse auto, clear")

    info = json.loads(get_ui_channel())
    base = info["baseUrl"]
    token = info["token"]
    headers = {"Authorization": f"Bearer {token}"}

    ds = httpx.get(urljoin(base, "/v1/dataset"), headers=headers).json()["dataset"]

    # Test missing limit - should fail
    page_req_no_limit = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "vars": ["price"],
    }
    r = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=page_req_no_limit)
    assert r.status_code == 400
    assert "limit is required" in r.json()["error"]["message"]

    # Test limit = 0 - should fail
    page_req_zero = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "limit": 0,
        "vars": ["price"],
    }
    r = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=page_req_zero)
    assert r.status_code == 400
    assert "limit must be > 0" in r.json()["error"]["message"]

    # Test limit = -1 - should fail
    page_req_negative = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "limit": -1,
        "vars": ["price"],
    }
    r = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=page_req_negative)
    assert r.status_code == 400
    assert "limit must be > 0" in r.json()["error"]["message"]

    # Test limit = null - should fail with "limit is required"
    page_req_null = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "limit": None,
        "vars": ["price"],
    }
    r = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=page_req_null)
    assert r.status_code == 400
    assert "limit is required" in r.json()["error"]["message"]

    # Test limit as string - should fail with type error
    page_req_string = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "limit": "not a number",
        "vars": ["price"],
    }
    r = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=page_req_string)
    assert r.status_code == 400
    assert "must be a valid integer" in r.json()["error"]["message"]

    # Test limit = 1 - should succeed
    page_req_valid = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "limit": 1,
        "vars": ["price"],
    }
    r = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=page_req_valid)
    assert r.status_code == 200
    assert r.json()["view"]["returned"] == 1

    # Test limit = 100 - should succeed
    page_req_100 = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "limit": 100,
        "vars": ["price"],
    }
    r = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=page_req_100)
    assert r.status_code == 200
    # auto dataset has 74 observations, so should return all of them
    assert r.json()["view"]["returned"] <= 100

    # Test limit exceeds max_limit (default 500) - should fail
    page_req_too_large = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "limit": 100_000,
        "vars": ["price"],
    }
    r = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=page_req_too_large)
    assert r.status_code == 400
    assert "limit must be <=" in r.json()["error"]["message"]


def test_ui_http_page_offset_validation():
    """Test that offset parameter is properly validated"""
    _run_command_sync("sysuse auto, clear")

    info = json.loads(get_ui_channel())
    base = info["baseUrl"]
    token = info["token"]
    headers = {"Authorization": f"Bearer {token}"}

    ds = httpx.get(urljoin(base, "/v1/dataset"), headers=headers).json()["dataset"]

    # Test offset = -1 - should fail
    page_req_negative_offset = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": -1,
        "limit": 10,
        "vars": ["price"],
    }
    r = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=page_req_negative_offset)
    assert r.status_code == 400
    assert "offset must be >=" in r.json()["error"]["message"]

    # Test offset = 0 - should succeed (default)
    page_req_zero_offset = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "limit": 10,
        "vars": ["price"],
    }
    r = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=page_req_zero_offset)
    assert r.status_code == 200

    # Test offset missing - should default to 0 and succeed
    page_req_no_offset = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "limit": 10,
        "vars": ["price"],
    }
    r = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=page_req_no_offset)
    assert r.status_code == 200
    assert r.json()["view"]["offset"] == 0

    # Test offset as invalid string - should fail
    page_req_invalid_offset = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": "not a number",
        "limit": 10,
        "vars": ["price"],
    }
    r = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=page_req_invalid_offset)
    assert r.status_code == 400
    assert "offset must be a valid integer" in r.json()["error"]["message"]


def test_ui_http_filter_price_less_than_5000():
    """Test that filtering for price < 5000 returns only rows where price < 5000"""
    # Load the auto dataset
    _run_command_sync("sysuse auto, clear")

    # Sort by price to make verification easier
    _run_command_sync("sort price")

    info = json.loads(get_ui_channel())
    base = info["baseUrl"]
    token = info["token"]
    headers = {"Authorization": f"Bearer {token}"}

    # Get dataset info
    ds = httpx.get(urljoin(base, "/v1/dataset"), headers=headers).json()["dataset"]

    # Create a view with filter: price < 5000
    view_req = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "filterExpr": "price < 5000",
    }
    view_resp = httpx.post(urljoin(base, "/v1/views"), headers=headers, json=view_req)
    assert view_resp.status_code == 200, f"Failed to create view: {view_resp.text}"

    view_payload = view_resp.json()
    view_id = view_payload["view"]["id"]
    filtered_n = view_payload["view"]["filteredN"]

    # The filtered count should be less than the total dataset
    assert filtered_n < ds["n"], f"Filtered count ({filtered_n}) should be less than total ({ds['n']})"
    assert filtered_n > 0, "Should have some rows with price < 5000"

    # Request all filtered rows with price variable
    page_req = {
        "offset": 0,
        "limit": 100,  # Get all rows
        "vars": ["price", "make"],
        "includeObsNo": True,
    }

    page_resp = httpx.post(urljoin(base, f"/v1/views/{view_id}/page"), headers=headers, json=page_req)
    assert page_resp.status_code == 200, f"Failed to get page: {page_resp.text}"

    page_data = page_resp.json()

    # Verify we got the expected number of rows
    assert page_data["view"]["returned"] == filtered_n, \
        f"Should return {filtered_n} rows, got {page_data['view']['returned']}"

    # Find the price column index
    vars_list = page_data["vars"]
    assert "price" in vars_list, "price should be in the returned variables"
    price_idx = vars_list.index("price")

    # Verify all returned rows have price < 5000
    rows = page_data["rows"]
    assert len(rows) == filtered_n, f"Should have {filtered_n} rows, got {len(rows)}"

    prices = []
    for i, row in enumerate(rows):
        price_value = row[price_idx]
        # Convert to float for comparison (might be string representation)
        try:
            price = float(price_value)
        except (ValueError, TypeError):
            # Handle missing values
            if price_value == "." or price_value is None:
                continue
            raise AssertionError(f"Row {i}: Invalid price value: {price_value!r}")

        prices.append(price)
        assert price < 5000, \
            f"Row {i}: price should be < 5000, got {price} (row data: {row})"

    # Verify we found at least some prices
    assert len(prices) > 0, "Should have found at least some non-missing prices"

    # Verify the data is sorted (since we sorted the dataset by price)
    # Note: The view should preserve the sort order
    for i in range(1, len(prices)):
        assert prices[i] >= prices[i-1], \
            f"Prices should be sorted: position {i-1} has {prices[i-1]}, position {i} has {prices[i]}"

    # Additional verification: fetch a page without the filter and ensure we have rows with price >= 5000
    unfiltered_page_req = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "limit": 100,
        "vars": ["price"],
    }

    unfiltered_resp = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=unfiltered_page_req)
    assert unfiltered_resp.status_code == 200

    unfiltered_data = unfiltered_resp.json()
    unfiltered_price_idx = unfiltered_data["vars"].index("price")

    # Check that the unfiltered dataset has at least some prices >= 5000
    has_expensive_cars = False
    for row in unfiltered_data["rows"]:
        price_value = row[unfiltered_price_idx]
        try:
            price = float(price_value)
            if price >= 5000:
                has_expensive_cars = True
                break
        except (ValueError, TypeError):
            continue

    assert has_expensive_cars, \
        "The full dataset should contain at least some cars with price >= 5000 to validate the filter is working"

    # Cleanup
    delete_resp = httpx.delete(urljoin(base, f"/v1/views/{view_id}"), headers=headers)
    assert delete_resp.status_code == 200

    print(f"✓ Filter test passed: {filtered_n} rows with price < 5000, all verified")
    print(f"✓ Price range in filtered data: {min(prices):.2f} to {max(prices):.2f}")


def test_ui_http_sorting_ascending():
    """Test sorting by a single variable in ascending order"""
    _run_command_sync("sysuse auto, clear")

    info = json.loads(get_ui_channel())
    base = info["baseUrl"]
    token = info["token"]
    headers = {"Authorization": f"Bearer {token}"}

    ds = httpx.get(urljoin(base, "/v1/dataset"), headers=headers).json()["dataset"]

    # Request page sorted by price ascending
    page_req = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "limit": 10,
        "vars": ["price", "make"],
        "sortBy": ["price"],  # or ["+price"]
    }

    resp = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=page_req)
    assert resp.status_code == 200, f"Failed: {resp.text}"

    data = resp.json()
    price_idx = data["vars"].index("price")
    prices = [float(row[price_idx]) for row in data["rows"]]

    # Verify prices are in ascending order
    for i in range(1, len(prices)):
        assert prices[i] >= prices[i-1], \
            f"Prices should be ascending: {prices[i-1]} > {prices[i]}"

    print(f"✓ Ascending sort test passed. Prices: {prices[0]:.0f} to {prices[-1]:.0f}")


def test_ui_http_sorting_descending():
    """Test sorting by a single variable in descending order"""
    _run_command_sync("sysuse auto, clear")

    info = json.loads(get_ui_channel())
    base = info["baseUrl"]
    token = info["token"]
    headers = {"Authorization": f"Bearer {token}"}

    ds = httpx.get(urljoin(base, "/v1/dataset"), headers=headers).json()["dataset"]

    # Request page sorted by price descending
    page_req = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "limit": 10,
        "vars": ["price", "make"],
        "sortBy": ["-price"],
    }

    resp = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=page_req)
    assert resp.status_code == 200, f"Failed: {resp.text}"

    data = resp.json()
    price_idx = data["vars"].index("price")
    prices = [float(row[price_idx]) for row in data["rows"]]

    # Verify prices are in descending order
    for i in range(1, len(prices)):
        assert prices[i] <= prices[i-1], \
            f"Prices should be descending: {prices[i-1]} < {prices[i]}"

    print(f"✓ Descending sort test passed. Prices: {prices[0]:.0f} to {prices[-1]:.0f}")


def test_ui_http_sorting_multiple_variables():
    """Test sorting by multiple variables"""
    _run_command_sync("sysuse auto, clear")

    info = json.loads(get_ui_channel())
    base = info["baseUrl"]
    token = info["token"]
    headers = {"Authorization": f"Bearer {token}"}

    ds = httpx.get(urljoin(base, "/v1/dataset"), headers=headers).json()["dataset"]

    # Sort by foreign (ascending), then price (descending)
    page_req = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "limit": 74,  # Get all rows
        "vars": ["foreign", "price", "make"],
        "sortBy": ["foreign", "-price"],
    }

    resp = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=page_req)
    assert resp.status_code == 200, f"Failed: {resp.text}"

    data = resp.json()
    foreign_idx = data["vars"].index("foreign")
    price_idx = data["vars"].index("price")

    # Extract foreign and price values
    rows_data = [(float(row[foreign_idx]), float(row[price_idx])) for row in data["rows"]]

    # Verify primary sort by foreign (ascending)
    for i in range(1, len(rows_data)):
        # If foreign values are different, current should be >= previous
        if rows_data[i][0] != rows_data[i-1][0]:
            assert rows_data[i][0] >= rows_data[i-1][0], \
                f"Foreign should be ascending: {rows_data[i-1][0]} > {rows_data[i][0]}"
        # If foreign values are the same, check price (descending)
        else:
            assert rows_data[i][1] <= rows_data[i-1][1], \
                f"Within same foreign group, price should be descending: {rows_data[i-1][1]} < {rows_data[i][1]}"

    print(f"✓ Multi-variable sort test passed")


def test_ui_http_sorting_with_filter():
    """Test that sorting works correctly with filtered views"""
    _run_command_sync("sysuse auto, clear")

    info = json.loads(get_ui_channel())
    base = info["baseUrl"]
    token = info["token"]
    headers = {"Authorization": f"Bearer {token}"}

    ds = httpx.get(urljoin(base, "/v1/dataset"), headers=headers).json()["dataset"]

    # Create a filtered view for price < 5000
    view_req = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "filterExpr": "price < 5000",
    }
    view_resp = httpx.post(urljoin(base, "/v1/views"), headers=headers, json=view_req)
    assert view_resp.status_code == 200
    view_id = view_resp.json()["view"]["id"]

    # Request sorted page from filtered view
    page_req = {
        "offset": 0,
        "limit": 50,
        "vars": ["price", "make"],
        "sortBy": ["-price"],  # Sort descending
    }

    resp = httpx.post(urljoin(base, f"/v1/views/{view_id}/page"), headers=headers, json=page_req)
    assert resp.status_code == 200, f"Failed: {resp.text}"

    data = resp.json()
    price_idx = data["vars"].index("price")
    prices = [float(row[price_idx]) for row in data["rows"]]

    # Verify all prices are < 5000 (filter still applied)
    for price in prices:
        assert price < 5000, f"All prices should be < 5000, got {price}"

    # Verify prices are in descending order (sort applied)
    for i in range(1, len(prices)):
        assert prices[i] <= prices[i-1], \
            f"Prices should be descending: {prices[i-1]} < {prices[i]}"

    # Cleanup
    httpx.delete(urljoin(base, f"/v1/views/{view_id}"), headers=headers)

    print(f"✓ Sort with filter test passed. Filtered & sorted {len(prices)} rows, range: ${prices[-1]:.0f} to ${prices[0]:.0f}")


def test_ui_http_sorting_invalid_variable():
    """Test that sorting with invalid variable returns proper error"""
    _run_command_sync("sysuse auto, clear")

    info = json.loads(get_ui_channel())
    base = info["baseUrl"]
    token = info["token"]
    headers = {"Authorization": f"Bearer {token}"}

    ds = httpx.get(urljoin(base, "/v1/dataset"), headers=headers).json()["dataset"]

    # Request with invalid sort variable
    page_req = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "limit": 10,
        "vars": ["price"],
        "sortBy": ["nonexistent_variable"],
    }

    resp = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=page_req)
    assert resp.status_code == 400
    assert "invalid" in resp.json()["error"]["message"].lower()

    print("✓ Invalid variable test passed")


def test_ui_http_sorting_invalid_format():
    """Test that invalid sortBy format returns proper error"""
    _run_command_sync("sysuse auto, clear")

    info = json.loads(get_ui_channel())
    base = info["baseUrl"]
    token = info["token"]
    headers = {"Authorization": f"Bearer {token}"}

    ds = httpx.get(urljoin(base, "/v1/dataset"), headers=headers).json()["dataset"]

    # Request with invalid sortBy format (not an array)
    page_req = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "limit": 10,
        "vars": ["price"],
        "sortBy": "price",  # Should be an array
    }

    resp = httpx.post(urljoin(base, "/v1/page"), headers=headers, json=page_req)
    assert resp.status_code == 400
    assert "array" in resp.json()["error"]["message"].lower()

    print("✓ Invalid format test passed")



from __future__ import annotations

import logging
from typing import Iterable, Any, Tuple

logger = logging.getLogger(__name__)

try:
    from mcp_stata import _native_ops as _native
except Exception:  # pragma: no cover - optional module
    _native = None


def argsort_numeric(
    columns: Iterable["numpy.ndarray"],
    descending: list[bool],
    nulls_last: list[bool],
) -> list[int] | None:
    if _native is None:
        return None
    cols = list(columns)
    if not cols:
        return []
    try:
        return _native.argsort_numeric(cols, descending, nulls_last)
    except Exception as e:
        logger.warning(f"Native numeric sort failed: {e}")
        return None


def argsort_mixed(
    columns: Iterable[object],
    is_string: list[bool],
    descending: list[bool],
    nulls_last: list[bool],
) -> list[int] | None:
    if _native is None:
        return None
    cols = list(columns)
    if not cols:
        return []
    try:
        return _native.argsort_mixed(cols, is_string, descending, nulls_last)
    except Exception as e:
        logger.warning(f"Native mixed sort failed: {e}")
        return None


def smcl_to_markdown(smcl_text: str) -> str | None:
    if _native is None:
        return None
    try:
        return _native.smcl_to_markdown(smcl_text)
    except Exception as e:
        logger.warning(f"Native SMCL conversion failed: {e}")
        return None


def fast_scan_log(smcl_content: str, rc_default: int) -> Tuple[str, str, int | None] | None:
    if _native is None:
        return None
    try:
        return _native.fast_scan_log(smcl_content, rc_default)
    except Exception as e:
        logger.warning(f"Native log scanning failed: {e}")
        return None


def compute_filter_indices(
    data_numeric: dict[str, "numpy.ndarray"],
    data_string: dict[str, list[str]],
    filter_json: str,
    row_count: int,
) -> list[int] | None:
    if _native is None:
        return None
    try:
        return _native.compute_filter_indices(
            data_numeric, 
            data_string, 
            filter_json, 
            row_count
        )
    except Exception as e:
        logger.warning(f"Native filtering failed: {e}")
        return None




import time
import re
import os
import numpy as np
from mcp_stata._native_ops import smcl_to_markdown as rust_smcl_to_markdown
from mcp_stata._native_ops import fast_scan_log as rust_fast_scan_log
from mcp_stata._native_ops import compute_filter_indices as rust_compute_filter_indices

# --- Task 1: SMCL to Markdown ---
def python_smcl_to_markdown(smcl_text: str) -> str:
    def _inline_to_markdown(text: str) -> str:
        def repl(match: re.Match) -> str:
            tag = match.group(1).lower()
            content = match.group(2) or ""
            if tag in ("bf", "strong"): return f"**{content}**"
            if tag in ("it", "em"): return f"*{content}*"
            if tag in ("cmd", "cmdab", "code", "inp", "input", "res", "err", "txt"): return f"`{content}`"
            return content
        text = re.sub(r"\{([a-zA-Z0-9_]+):([^}]*)\}", repl, text)
        text = re.sub(r"\{[^}]*\}", "", text)
        return text

    if not smcl_text: return ""
    lines = smcl_text.splitlines()
    body_parts = []
    title = None
    for raw in lines:
        line = raw.strip()
        if not line or line == "{smcl}": continue
        if line.startswith("{title:"):
            title = line[len("{title:") :].rstrip("}")
            continue
        line = line.replace("{p_end}", "")
        line = re.sub(r"\{p[^}]*\}", "", line)
        body_parts.append(_inline_to_markdown(line))
    
    res = ""
    if title: res += f"## {title}\n\n"
    res += "\n".join(part for part in body_parts if part)
    return res

# --- Task 2: Log Scanning ---
def python_fast_scan_log(smcl_content: str, rc_default: int):
    rc = None
    matches = list(re.finditer(r'\{search r\((\d+)\)', smcl_content))
    if matches:
        rc = int(matches[-1].group(1))
    if rc is None:
        matches = list(re.finditer(r'(?<!\w)r\((\d+)\);?', smcl_content))
        if matches:
            rc = int(matches[-1].group(1))
    
    lines = smcl_content.splitlines()
    error_msg = f"Stata error r({rc or rc_default})"
    error_start_idx = -1
    for i in range(len(lines) - 1, -1, -1):
        if '{err}' in lines[i]:
            error_start_idx = i
            j = i
            err_lines = []
            while j >= 0 and '{err}' in lines[j]:
                cleaned = re.sub(r'\{[^}]*\}', '', lines[j]).strip()
                if cleaned: err_lines.insert(0, cleaned)
                j -= 1
            if err_lines: error_msg = " ".join(err_lines)
            break
    
    context_start = max(0, error_start_idx - 5) if error_start_idx >= 0 else max(0, len(lines) - 30)
    context = "\n".join(lines[context_start:])
    return error_msg, context, rc

# --- Task 3: Filtering ---
def python_compute_filter(expr: str, names: list, columns: list):
    # Simplified simulation of the eval loop
    indices = []
    row_count = len(columns[0])
    # Pre-compile
    code = compile(expr, "<string>", "eval")
    for i in range(row_count):
        env = {names[j]: columns[j][i] for j in range(len(names))}
        if eval(code, {"__builtins__": {}}, env):
            indices.append(i)
    return indices

def run_benchmarks():
    print("--- Rust vs Python Benchmarks ---")
    
    # Benchmark SMCL
    smcl_sample = "{smcl}\n{title:Help for regress}\n{p 4 4 2}\n{bf:regress} fits a model of {it:depvar} on {it:indepvars}.\n{p_end}\n" * 1000
    t0 = time.perf_counter()
    py_smcl = python_smcl_to_markdown(smcl_sample)
    t1 = time.perf_counter()
    rust_smcl = rust_smcl_to_markdown(smcl_sample)
    t2 = time.perf_counter()
    print(f"SMCL (4000 lines): Python={t1-t0:.4f}s, Rust={t2-t1:.4f}s (Ratio: {(t1-t0)/(t2-t1):.1f}x)")

    # Benchmark Log Scan
    log_sample = "some output\n" * 50000 + "{err}variable not found\n{err}r(111);\n{search r(111), ...}\n"
    t0 = time.perf_counter()
    py_log = python_fast_scan_log(log_sample, 0)
    t1 = time.perf_counter()
    rust_log = rust_fast_scan_log(log_sample, 0)
    t2 = time.perf_counter()
    print(f"Log Scan (50000 lines): Python={t1-t0:.4f}s, Rust={t2-t1:.4f}s (Ratio: {(t1-t0)/(t2-t1):.1f}x)")

    # Benchmark Filtering
    N = 100000
    names = ["price", "mpg"]
    price = np.random.rand(N) * 10000
    mpg = np.random.rand(N) * 30
    expr = "(price > 5000) && (mpg < 20)"
    
    # Rust needs specific types
    t0 = time.perf_counter()
    # Mocking what the client would do: pass numpy arrays directly
    rust_res = rust_compute_filter_indices(expr, names, [price, mpg], [False, False])
    t1 = time.perf_counter()
    print(f"Filtering ({N} rows): Rust={t1-t0:.4f}s")
    
    # Python is too slow for 100k rows in a real eval loop helper, let's do 10k for comparison
    N_small = 10000
    price_s = price[:N_small]
    mpg_s = mpg[:N_small]
    expr_py = "(price > 5000) and (mpg < 20)"
    t0 = time.perf_counter()
    py_res = python_compute_filter(expr_py, names, [price_s, mpg_s])
    t1 = time.perf_counter()
    print(f"Filtering ({N_small} rows): Python={t1-t0:.4f}s")
    print(f"Extrapolated Python for {N} rows: {(t1-t0)*10:.4f}s (Ratio: {((t1-t0)*10)/(rust_res_t := (t1-t0 if 'rust_res_t' not in locals() else rust_res_t)):.1f}x)")
    # Correcting ratio calculation
    rust_N = (t1_rust := time.perf_counter()) - (t0_rust := time.perf_counter()) # placeholder
    t0_rust = time.perf_counter()
    rust_compute_filter_indices(expr, names, [price, mpg], [False, False])
    t1_rust = time.perf_counter()
    rust_time = t1_rust - t0_rust
    print(f"Actual Ratio for Filtering: {((t1-t0)*10)/rust_time:.1f}x")

if __name__ == "__main__":
    run_benchmarks()



import json
from pathlib import Path
import tempfile

from mcp_stata.server import read_log, find_in_log


def _write_temp_log(content: str) -> Path:
    temp = tempfile.NamedTemporaryFile(mode="w", delete=False, suffix=".log", encoding="utf-8")
    try:
        temp.write(content)
        temp.flush()
    finally:
        temp.close()
    return Path(temp.name)


def test_read_log_basic():
    log_path = _write_temp_log("line1\nline2\nline3\n")
    try:
        payload = json.loads(read_log(str(log_path)))
        assert payload["path"] == str(log_path)
        assert payload["offset"] == 0
        assert "line2" in payload["data"]
        assert payload["next_offset"] > 0
    finally:
        log_path.unlink(missing_ok=True)


def test_read_log_offset_and_max_bytes():
    content = "header\nalpha\nbeta\ngamma\n"
    log_path = _write_temp_log(content)
    try:
        first = json.loads(read_log(str(log_path), offset=0, max_bytes=8))
        assert first["data"]
        next_offset = first["next_offset"]
        second = json.loads(read_log(str(log_path), offset=next_offset, max_bytes=1024))
        assert second["offset"] == next_offset
        assert "gamma" in second["data"]
    finally:
        log_path.unlink(missing_ok=True)


def test_read_log_missing_file():
    payload = json.loads(read_log("/tmp/mcp_stata_missing.log"))
    assert payload["data"] == ""
    assert payload["next_offset"] == payload["offset"]


def test_find_in_log_literal_and_context():
    log_path = _write_temp_log("one\ntwo\nthree\nfour\n")
    try:
        payload = json.loads(find_in_log(str(log_path), "three", before=1, after=1))
        assert payload["matches"]
        context = payload["matches"][0]["context"]
        assert context == ["two", "three", "four"]
    finally:
        log_path.unlink(missing_ok=True)


def test_find_in_log_regex_case_sensitive():
    log_path = _write_temp_log("Alpha\nbeta\nALPHA\n")
    try:
        insensitive = json.loads(find_in_log(str(log_path), "alpha", case_sensitive=False))
        assert len(insensitive["matches"]) == 2
        sensitive = json.loads(find_in_log(str(log_path), r"^Alpha$", regex=True, case_sensitive=True))
        assert len(sensitive["matches"]) == 1
    finally:
        log_path.unlink(missing_ok=True)


def test_find_in_log_start_offset_and_max_matches():
    log_path = _write_temp_log("hit\nmiss\nhit\nmiss\nhit\n")
    try:
        first_pass = json.loads(find_in_log(str(log_path), "hit", max_matches=1, max_bytes=8))
        assert len(first_pass["matches"]) == 1
        next_offset = first_pass["next_offset"]
        second_pass = json.loads(find_in_log(str(log_path), "hit", start_offset=next_offset, max_matches=2))
        assert len(second_pass["matches"]) >= 1
    finally:
        log_path.unlink(missing_ok=True)




import os
import sys

import pytest
import stata_setup

# Skip entirely on non-Windows platforms since Stata COM setup is Windows-only.
pytestmark = pytest.mark.skipif(sys.platform != "win32", reason="Stata setup test runs only on Windows")

stata_exec_path = r"C:\Program Files\StataNow19\StataSE-64.exe"
edition = "se"

try:
    print(f"Attempting to run stata_setup.config with path: {os.path.dirname(stata_exec_path)} and edition: {edition}")
    stata_setup.config(os.path.dirname(stata_exec_path), edition)
    print("stata_setup.config succeeded")
except Exception as e:
    print(f"stata_setup.config failed: {e}")



import pytest
from mcp_stata.stata_client import StataClient

class TestRCParsingUnit:
    @pytest.fixture
    def client(self):
        return StataClient()

    def test_parse_rc_from_smcl_search_tag(self, client):
        smcl = "{txt}{search r(111), local:r(111);}"
        assert client._parse_rc_from_smcl(smcl) == 111

    def test_parse_rc_from_smcl_standalone(self, client):
        smcl = "some output\nr(601);\n"
        assert client._parse_rc_from_smcl(smcl) == 601

    def test_parse_rc_from_smcl_false_positive(self, client):
        # char(10) ends in r(10) but should not match without semicolon or search tag
        smcl = "{txt}241{com}.         local NL = char(10)"
        assert client._parse_rc_from_smcl(smcl) is None

    def test_parse_rc_from_smcl_multi_line(self, client):
        smcl = """
{res}{err}variable {bf}compl_gloves{sf} not found
{txt}{search r(111), local:r(111);}
"""
        assert client._parse_rc_from_smcl(smcl) == 111

    def test_parse_rc_from_text_search_pattern(self, client):
        text = "search r(111), local:r(111);"
        assert client._parse_rc_from_text(text) == 111

    def test_parse_rc_from_text_standalone(self, client):
        text = "error happened\nr(198);"
        assert client._parse_rc_from_text(text) == 198

    def test_parse_rc_from_text_false_positive(self, client):
        text = "local NL = char(10)"
        assert client._parse_rc_from_text(text) is None

    def test_parse_rc_from_smcl_complex_boundary(self, client):
        # Should match if preceded by space or start of line
        assert client._parse_rc_from_smcl(" r(123);") == 123
        assert client._parse_rc_from_smcl("\nr(123);") == 123
        # Should NOT match if preceded by word character (like char)
        assert client._parse_rc_from_smcl("char(10);") is None



"""
Tests for enhanced pystata integration in graph detection using actual implementation.
"""

import pytest
import os

# Configure Stata before importing sfi-dependent modules
import stata_setup
from conftest import configure_stata_for_tests

try:
    stata_dir, stata_flavor = configure_stata_for_tests()
    stata_setup.config(stata_dir, stata_flavor)
except (FileNotFoundError, PermissionError) as e:
    pytest.skip(f"Stata not found or not executable: {e}", allow_module_level=True)

from mcp_stata.graph_detector import GraphCreationDetector, StreamingGraphCache


# Mark all tests in this module as requiring Stata
pytestmark = pytest.mark.requires_stata


class TestPystataGraphIntegration:
    """Test pystata integration for graph detection using real Stata."""
    
    @pytest.fixture
    def detector_with_real_client(self, client):
        """Create detector with shared StataClient."""
        return GraphCreationDetector(stata_client=client)
    
    def test_detector_init_with_client(self, detector_with_real_client):
        """Test GraphCreationDetector initialization with real StataClient."""
        detector = detector_with_real_client
        assert detector._stata_client is not None
        assert hasattr(detector._stata_client, 'stata')
    
    def test_detect_graphs_via_pystata_new_graphs(self, detector_with_real_client):
        """Test detecting new graphs via pystata."""
        detector = detector_with_real_client
        
        # Clear existing graphs
        detector._stata_client.stata.run("clear", quietly=True)
        
        # Create a graph
        detector._stata_client.stata.run("sysuse auto, clear", quietly=True)
        detector._stata_client.stata.run("scatter price mpg, name(NewGraph)", quietly=True)
        
        # Detect new graphs
        detected = detector._detect_graphs_via_pystata()
        
        assert "NewGraph" in detected, f"Should detect NewGraph, found: {detected}"
        
        # Clean up
        detector._stata_client.stata.run("graph drop NewGraph", quietly=True)
        detector._stata_client.stata.run("clear", quietly=True)
    
    def test_detect_graphs_via_pystata_modified_graphs(self, detector_with_real_client):
        """Test detecting modified graphs via pystata."""
        detector = detector_with_real_client
        
        # Create initial graph
        detector._stata_client.stata.run("sysuse auto, clear", quietly=True)
        detector._stata_client.stata.run("scatter price mpg, name(Graph1)", quietly=True)
        
        # Get initial state
        initial_state = detector._get_graph_state_from_pystata()
        
        # Modify the graph
        detector._stata_client.stata.run("graph drop Graph1", quietly=True)
        detector._stata_client.stata.run("scatter price weight, name(Graph1)", quietly=True)
        
        # Detect modified graphs
        detected = detector._detect_graphs_via_pystata()
        
        assert "Graph1" in detected, f"Should detect modified Graph1, found: {detected}"
        
        # Clean up
        detector._stata_client.stata.run("graph drop Graph1", quietly=True)
        detector._stata_client.stata.run("clear", quietly=True)
    
    def test_get_current_graphs_from_pystata_with_client(self, detector_with_real_client):
        """Test getting current graphs from pystata with real StataClient."""
        detector = detector_with_real_client
        
        # Create a graph
        detector._stata_client.stata.run("sysuse auto, clear", quietly=True)
        detector._stata_client.stata.run("scatter price mpg, name(CurrentTest)", quietly=True)
        
        # Get current graphs
        graphs = detector._get_current_graphs_from_pystata()
        
        assert "CurrentTest" in graphs, f"Should detect CurrentTest, found: {graphs}"
        
        # Clean up
        detector._stata_client.stata.run("graph drop CurrentTest", quietly=True)
        detector._stata_client.stata.run("clear", quietly=True)
    
    def test_get_graph_state_from_pystata(self, detector_with_real_client):
        """Test getting graph state from pystata."""
        detector = detector_with_real_client
        
        # Create a graph
        detector._stata_client.stata.run("sysuse auto, clear", quietly=True)
        detector._stata_client.stata.run("scatter price mpg, name(StateTest)", quietly=True)
        
        # Get graph state
        state = detector._get_graph_state_from_pystata()
        
        assert "StateTest" in state, "Should contain StateTest"
        assert state["StateTest"]["name"] == "StateTest", "Should have correct name"
        assert state["StateTest"]["exists"] == True, "Should show graph exists"
        
        # Clean up
        detector._stata_client.stata.run("graph drop StateTest", quietly=True)
        detector._stata_client.stata.run("clear", quietly=True)
    
    def test_sfi_graph_detection_integration(self, detector_with_real_client):
        """Test SFI graph detection integration with real Stata."""
        detector = detector_with_real_client
        
        # Test SFI detection with real Stata
        detected = detector._detect_graphs_via_pystata()
        assert isinstance(detected, list)
        
        # Test state tracking
        state = detector._get_graph_state_from_pystata()
        assert isinstance(state, dict)
        
        # Clean up
        detector._stata_client.stata.run("clear", quietly=True)
    
    def test_sfi_detection_integration_complete(self, detector_with_real_client):
        """Test complete SFI detection integration with real Stata."""
        detector = detector_with_real_client
        
        # Clear existing graphs
        detector._stata_client.stata.run("clear", quietly=True)
        
        # Create a graph
        detector._stata_client.stata.run("sysuse auto, clear", quietly=True)
        output = detector._stata_client.stata.run("scatter price mpg, name(IntegrationTest)", quietly=True)
        
        # Detect graphs using SFI
        detected = detector._detect_graphs_via_pystata()
        
        assert "IntegrationTest" in detected, f"Should detect IntegrationTest, found: {detected}"
        
        # Clean up
        detector._stata_client.stata.run("graph drop IntegrationTest", quietly=True)
        detector._stata_client.stata.run("clear", quietly=True)
    
        
    @pytest.mark.asyncio
    async def test_cache_detected_graphs_with_pystata(self, client):
        """Test enhanced caching with pystata integration."""
        cache = StreamingGraphCache(client, auto_cache=True)
        
        # Create real graphs first
        client.stata.run("sysuse auto, clear", quietly=True)
        client.stata.run("scatter price mpg, name(Graph1)", quietly=True)
        client.stata.run("histogram price, name(Graph2)", quietly=True)
        
        # Add graphs to cache queue
        cache._graphs_to_cache = ['Graph1', 'Graph2']
        
        # Test with real graphs
        cached_graphs = await cache.cache_detected_graphs_with_pystata()
        
        # Should cache graphs from queue and pystata detection
        assert len(cached_graphs) >= 0
        
        # Clean up
        try:
            client.stata.run("graph drop Graph1 Graph2", quietly=True)
        except SystemError:
            pass
        client.stata.run("clear", quietly=True)
    
    @pytest.mark.asyncio
    async def test_cache_detected_graphs_fallback(self, client):
        """Test fallback to original method when pystata method not available."""
        cache = StreamingGraphCache(client, auto_cache=True)
        
        # Create a real graph first
        client.stata.run("sysuse auto, clear", quietly=True)
        client.stata.run("scatter price mpg, name(Graph1)", quietly=True)
        
        # Add graph to cache queue
        cache._graphs_to_cache = ['Graph1']
        
        cached_graphs = await cache.cache_detected_graphs()
        
        # Should cache the real graph
        assert len(cached_graphs) >= 0
        
        # Clean up
        try:
            client.stata.run("graph drop Graph1", quietly=True)
        except SystemError:
            pass
        client.stata.run("clear", quietly=True)

    def test_svg_export_integration(self, client):
        """Test SVG export integration with real Stata."""
        stata = client.stata
        
        # Create a graph
        stata.run("sysuse auto, clear", quietly=True)
        stata.run("scatter price mpg, name(SVGTestGraph)", quietly=True)
        
        # Export to SVG
        svg_path = os.path.abspath("SVGTestGraph.svg")
        stata.run(f"graph export {svg_path}, as(svg) replace", quietly=True)
        
        # Check if SVG file was created
        assert os.path.isfile(svg_path), "SVG file should be created"
        
        # Clean up
        stata.run("graph drop SVGTestGraph", quietly=True)
        stata.run("clear", quietly=True)
        if os.path.isfile(svg_path):
            os.remove(svg_path)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



import anyio
import pytest
import json
from pathlib import Path


pytestmark = pytest.mark.requires_stata


def test_run_command_streaming_emits_log_and_progress(client):

    logs: list[str] = []
    progress: list[tuple[float, float | None, str | None]] = []

    async def notify_log(chunk: str) -> None:
        logs.append(chunk)

    async def notify_progress(p: float, total: float | None, msg: str | None) -> None:
        progress.append((p, total, msg))

    async def main():
        res = await client.run_command_streaming(
            "display 5+5",
            notify_log=notify_log,
            notify_progress=notify_progress,
            echo=True,
        )
        assert res.rc == 0
        assert res.stdout == ""
        assert res.log_path is not None
        assert Path(res.log_path).exists()

        # Output should be in the log file.
        text = Path(res.log_path).read_text(encoding="utf-8", errors="replace")
        assert "10" in text

    anyio.run(main)


def test_run_command_streaming_with_cwd_can_do_relative_file(tmp_path, client):
    project = tmp_path / "proj_cmd"
    project.mkdir()
    dofile = project / "rel.do"
    dofile.write_text('display "cmd-ok"\n')

    logs: list[str] = []

    async def notify_log(chunk: str) -> None:
        logs.append(chunk)

    async def main():
        res = await client.run_command_streaming(
            'do "rel.do"',
            notify_log=notify_log,
            notify_progress=None,
            echo=True,
            cwd=str(project),
        )
        assert res.success is True
        assert res.rc == 0
        assert res.log_path is not None
        text = Path(res.log_path).read_text(encoding="utf-8", errors="replace")
        assert "cmd-ok" in text

    anyio.run(main)

    # Should emit at least one log output event (log_path)
    assert len(logs) > 0
    payload = json.loads(logs[0])
    assert payload.get("event") == "log_path"
    assert Path(payload.get("path", "")).exists()


def test_run_do_file_streaming_progress_inference(tmp_path, client):
    dofile = tmp_path / "stream_test.do"
    dofile.write_text('display "a"\ndisplay "b"\n')

    logs: list[str] = []
    progress: list[tuple[float, float | None, str | None]] = []

    async def notify_log(chunk: str) -> None:
        logs.append(chunk)

    async def notify_progress(p: float, total: float | None, msg: str | None) -> None:
        progress.append((p, total, msg))

    async def main():
        res = await client.run_do_file_streaming(
            str(dofile),
            notify_log=notify_log,
            notify_progress=notify_progress,
            echo=True,
        )
        assert res.rc == 0
        assert res.stdout == ""
        assert res.log_path is not None
        assert Path(res.log_path).exists()

        text = Path(res.log_path).read_text(encoding="utf-8", errors="replace")
        assert "a" in text
        assert "b" in text

    anyio.run(main)

    assert len(logs) > 0
    payload = json.loads(logs[0])
    assert payload.get("event") == "log_path"
    assert Path(payload.get("path", "")).exists()
    # At minimum we should have the initial progress message, and often inferred updates.
    assert len(progress) >= 1


def test_run_do_file_streaming_with_cwd_and_relative_paths(tmp_path, client):
    project = tmp_path / "proj"
    project.mkdir()
    child = project / "child.do"
    child.write_text('display "child-ok"\n')
    parent = project / "parent.do"
    parent.write_text('do "child.do"\ndisplay "parent-ok"\n')

    logs: list[str] = []

    async def notify_log(chunk: str) -> None:
        logs.append(chunk)

    async def main():
        res = await client.run_do_file_streaming(
            "parent.do",
            notify_log=notify_log,
            notify_progress=None,
            echo=True,
            cwd=str(project),
        )
        assert res.success is True
        assert res.rc == 0
        assert res.log_path is not None
        text = Path(res.log_path).read_text(encoding="utf-8", errors="replace")
        assert "child-ok" in text
        assert "parent-ok" in text

    anyio.run(main)

if __name__ == "__main__":
    pytest.main([__file__, "-v"])


import platform

from mcp_stata import discovery


def test_discovery_cli_main_success(monkeypatch, tmp_path, capsys):
    # Create a fake executable path and point STATA_PATH at it
    fake_bin = tmp_path / "Program Files" / "Stata18" / "StataMP-64.exe"
    fake_bin.parent.mkdir(parents=True, exist_ok=True)
    fake_bin.write_text("stub")
    monkeypatch.setenv("STATA_PATH", str(fake_bin))
    monkeypatch.setattr(platform, "system", lambda: "Windows")

    rc = discovery.main()
    captured = capsys.readouterr().out
    assert rc == 0
    assert "Stata executable" in captured
    assert str(fake_bin) in captured
    assert "mp" in captured


def test_discovery_cli_main_failure(monkeypatch, capsys):
    # Force failure by making discovery search empty candidates
    monkeypatch.delenv("STATA_PATH", raising=False)
    
    # Mock _detect_system directly to bypass os.name check
    monkeypatch.setattr(discovery, "_detect_system", lambda: "NowhereOS")
    
    rc = discovery.main()
    captured = capsys.readouterr().out
    assert rc == 1
    assert "Discovery failed" in captured



import pytest
from unittest.mock import MagicMock, patch
import pandas as pd
import pyarrow as pa
import sys

from mcp_stata.stata_client import StataClient

class TestArrowUnit:
    @pytest.fixture
    def client(self):
        # Create a client with mocked internals
        client = StataClient()
        client._initialized = True
        client.stata = MagicMock()
        return client

    def test_get_arrow_stream_basic(self, client):
        # Setup mocks
        mock_data_ns = MagicMock()
        mock_data_ns.getObsTotal.return_value = 10
        mock_data_ns.getVarCount.return_value = 2
        
        # We need to mock sfi.Data being imported inside methods
        with patch.dict(sys.modules, {"sfi": MagicMock(Data=mock_data_ns)}):
            # Mock Data.get to return valid data (nested list of rows)
            mock_data_ns.get.return_value = [[1, "a"], [2, "b"]]
            # Mock get_dataset_state to return non-empty
            with patch.object(client, "get_dataset_state", return_value={"n": 10, "k": 2}):
                with patch.object(client, "_get_var_index_map", return_value={"v1": 0, "v2": 1}):
                    # Mock pystata response
                    df_mock = pd.DataFrame({"v1": [1, 2], "v2": ["a", "b"]})
                    client.stata.pdataframe_from_data.return_value = df_mock
                    
                    # Execute
                    arrow_bytes = client.get_arrow_stream(
                        offset=0, 
                        limit=2, 
                        vars=["v1", "v2"], 
                        include_obs_no=False
                    )

                    # Verify calling arguments
                    # Note: we can't easily check 'obs' arg if it was a list object, 
                    # but we can check it was called.
                    mock_data_ns.get.assert_called_once()
                    
                    # Verify output is valid Arrow stream
                    reader = pa.ipc.open_stream(arrow_bytes)
                    table = reader.read_all()
                    
                    assert table.num_rows == 2
                    assert table.num_columns == 2
                    assert table.column_names == ["v1", "v2"]
                    assert table["v1"].to_pylist() == [1, 2]
                    assert table["v2"].to_pylist() == ["a", "b"]

    def test_get_arrow_stream_with_obs_no(self, client):
        mock_data_ns = MagicMock()
        mock_data_ns.get.return_value = [[10.5]]
        
        with patch.dict(sys.modules, {"sfi": MagicMock(Data=mock_data_ns)}):
            with patch.object(client, "get_dataset_state", return_value={"n": 10, "k": 1}):
                with patch.object(client, "_get_var_index_map", return_value={"v1": 0}):
                    df_mock = pd.DataFrame({"v1": [10.5]})
                    client.stata.pdataframe_from_data.return_value = df_mock
                    
                    arrow_bytes = client.get_arrow_stream(
                        offset=5, 
                        limit=1, 
                        vars=["v1"], 
                        include_obs_no=True
                    )
                    
                    reader = pa.ipc.open_stream(arrow_bytes)
                    table = reader.read_all()
                    
                    assert table.num_rows == 1
                    # Should have _n + v1
                    assert table.column_names == ["_n", "v1"]
                    # _n should be offset (5) + 1 = 6
                    assert table["_n"].to_pylist() == [6]

    def test_get_arrow_stream_empty(self, client):
        mock_data_ns = MagicMock()
        
        with patch.dict(sys.modules, {"sfi": MagicMock(Data=mock_data_ns)}):
            with patch.object(client, "get_dataset_state", return_value={"n": 10, "k": 1}):
                with patch.object(client, "_get_var_index_map", return_value={"v1": 0}):
                    # Mock return for empty list (pystata typically returns empty df)
                    client.stata.pdataframe_from_data.return_value = pd.DataFrame(columns=["v1"])
                    
                    # Request out of bounds
                    arrow_bytes = client.get_arrow_stream(
                        offset=100, 
                        limit=10, 
                        vars=["v1"], 
                        include_obs_no=False
                    )
                    
                    reader = pa.ipc.open_stream(arrow_bytes)
                    try:
                        table = reader.read_all()
                        assert table.num_rows == 0
                        assert table.column_names == ["v1"]
                    except pa.ArrowInvalid:
                        # Some versions might fail reading empty stream without batches
                        # but we wrote a table, so it should have schema even if empty.
                        pass

from mcp_stata.ui_http import handle_arrow_request, HTTPError, UIChannelManager

class TestArrowHandlerUnit:
    @pytest.fixture
    def manager(self):
        manager = MagicMock(spec=UIChannelManager)
        manager.limits.return_value = (500, 200, 500, 1_000_000)
        manager.current_dataset_id.return_value = "test_id"
        manager._max_arrow_limit = 1_000_000
        manager._client = MagicMock()
        return manager

    def test_handle_arrow_request_missing_limit(self, manager):
        body = {
            "datasetId": "test_id",
            "frame": "default",
            "offset": 0,
            "vars": [],
        }
        with pytest.raises(HTTPError) as exc_info:
            handle_arrow_request(manager, body, view_id=None)
        assert exc_info.value.status == 400
        assert "limit is required" in exc_info.value.message

    def test_handle_arrow_request_limit_too_large(self, manager):
        manager._max_arrow_limit = 1000
        body = {
            "datasetId": "test_id",
            "frame": "default",
            "offset": 0,
            "limit": 5000,
            "vars": [],
        }
        with pytest.raises(HTTPError) as exc_info:
            handle_arrow_request(manager, body, view_id=None)
        assert exc_info.value.status == 400
        assert "request_too_large" in exc_info.value.code
        assert "limit must be <= 1000" in exc_info.value.message

    def test_handle_arrow_request_dataset_changed(self, manager):
        manager.current_dataset_id.return_value = "new_id"
        body = {
            "datasetId": "old_id",
            "frame": "default",
            "offset": 0,
            "limit": 10,
            "vars": [],
        }
        with pytest.raises(HTTPError) as exc_info:
            handle_arrow_request(manager, body, view_id=None)
        assert exc_info.value.status == 409
        assert "dataset_changed" in exc_info.value.code

    def test_handle_arrow_request_valid(self, manager):
        manager._client.get_arrow_stream.return_value = b"arrow_data"
        manager._normalize_sort_spec.return_value = ("+v1",)
        manager._get_cached_sort_indices.return_value = [2, 1, 0]
        body = {
            "datasetId": "test_id",
            "frame": "default",
            "offset": 10,
            "limit": 50,
            "vars": ["v1", "v2"],
            "includeObsNo": True,
            "sortBy": ["v1"]
        }
        
        result = handle_arrow_request(manager, body, view_id=None)
        
        assert result == b"arrow_data"
        manager._client.get_arrow_stream.assert_called_once_with(
            offset=10,
            limit=50,
            vars=["v1", "v2"],
            include_obs_no=True,
            obs_indices=[2, 1, 0]
        )



import json
from urllib.parse import urljoin
import io

import anyio
import httpx
import pytest
import pyarrow as pa

from mcp_stata.server import get_ui_channel, run_command

pytestmark = [pytest.mark.requires_stata, pytest.mark.integration]

def _run_command_sync(code: str) -> str:
    async def _main() -> str:
        return await run_command(code)
    return anyio.run(_main)

def test_ui_http_arrow_basic():
    """Verify that /v1/arrow returns a valid Arrow stream with expected data"""
    _run_command_sync("sysuse auto, clear")

    info = json.loads(get_ui_channel())
    base = info["baseUrl"]
    token = info["token"]
    headers = {"Authorization": f"Bearer {token}"}

    # Get dataset info to get the ID
    ds = httpx.get(urljoin(base, "/v1/dataset"), headers=headers).json()["dataset"]

    arrow_req = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "limit": 10,
        "vars": ["price", "mpg", "make"],
        "includeObsNo": True
    }

    # Request the arrow stream
    r = httpx.post(urljoin(base, "/v1/arrow"), headers=headers, json=arrow_req)
    assert r.status_code == 200
    assert r.headers["Content-Type"] == "application/vnd.apache.arrow.stream"
    
    # Binary body
    data_bytes = r.content
    assert len(data_bytes) > 0
    
    # Parse with pyarrow
    with pa.ipc.open_stream(io.BytesIO(data_bytes)) as reader:
        table = reader.read_all()
    
    assert table.num_rows == 10
    assert "_n" in table.column_names
    assert "price" in table.column_names
    assert "mpg" in table.column_names
    assert "make" in table.column_names
    
    # Verify some values (auto.dta first row is AMC Concord, price 4099, mpg 22)
    # price and mpg are likely float64 or int64 in Arrow
    assert table["price"][0].as_py() == 4099
    assert table["mpg"][0].as_py() == 22
    assert table["_n"][0].as_py() == 1

def test_ui_http_arrow_sorting():
    """Verify that /v1/arrow respects the sortBy parameter"""
    _run_command_sync("sysuse auto, clear")

    info = json.loads(get_ui_channel())
    base = info["baseUrl"]
    token = info["token"]
    headers = {"Authorization": f"Bearer {token}"}

    ds = httpx.get(urljoin(base, "/v1/dataset"), headers=headers).json()["dataset"]

    # Sort DESCENDING by price
    arrow_req = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "limit": 5,
        "vars": ["price"],
        "sortBy": ["-price"]
    }

    r = httpx.post(urljoin(base, "/v1/arrow"), headers=headers, json=arrow_req)
    assert r.status_code == 200
    
    with pa.ipc.open_stream(io.BytesIO(r.content)) as reader:
        table = reader.read_all()
    
    prices = table["price"].to_pylist()
    assert len(prices) == 5
    # Verify they are descending
    for i in range(1, len(prices)):
        assert prices[i] <= prices[i-1]

def test_ui_http_arrow_large_limit():
    """Verify that /v1/arrow supports much larger limits than /v1/page"""
    _run_command_sync("sysuse auto, clear")

    info = json.loads(get_ui_channel())
    base = info["baseUrl"]
    token = info["token"]
    headers = {"Authorization": f"Bearer {token}"}

    ds = httpx.get(urljoin(base, "/v1/dataset"), headers=headers).json()["dataset"]

    # Limit of 1000 (auto only has 74, but this verifies it doesn't fail due to UI_LIMIT=500)
    arrow_req = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "offset": 0,
        "limit": 1000,
        "vars": ["price"]
    }

    r = httpx.post(urljoin(base, "/v1/arrow"), headers=headers, json=arrow_req)
    assert r.status_code == 200
    
    with pa.ipc.open_stream(io.BytesIO(r.content)) as reader:
        table = reader.read_all()
    
    assert table.num_rows == 74 # Full auto dataset

def test_ui_http_arrow_with_view():
    """Verify that /v1/arrow works within a filtered view"""
    _run_command_sync("sysuse auto, clear")

    info = json.loads(get_ui_channel())
    base = info["baseUrl"]
    token = info["token"]
    headers = {"Authorization": f"Bearer {token}"}

    ds = httpx.get(urljoin(base, "/v1/dataset"), headers=headers).json()["dataset"]

    # Create a view for foreign cars
    view_req = {
        "datasetId": ds["id"],
        "frame": ds.get("frame", "default"),
        "filterExpr": "foreign == 1"
    }
    view_resp = httpx.post(urljoin(base, "/v1/views"), headers=headers, json=view_req)
    view_id = view_resp.json()["view"]["id"]
    filtered_n = view_resp.json()["view"]["filteredN"]

    # Request arrow from this view
    arrow_req = {
        "offset": 0,
        "limit": 50,
        "vars": ["price", "foreign"]
    }
    r = httpx.post(urljoin(base, f"/v1/views/{view_id}/arrow"), headers=headers, json=arrow_req)
    assert r.status_code == 200
    
    with pa.ipc.open_stream(io.BytesIO(r.content)) as reader:
        table = reader.read_all()
    
    assert table.num_rows == filtered_n
    # foreign should be 1 for all
    for val in table["foreign"].to_pylist():
        assert val == 1




"""
Test TTL cache functionality for list_graphs() method.
"""

import pytest
import threading
import time
from conftest import configure_stata_for_tests
import stata_setup
pytestmark = pytest.mark.requires_stata

try:
    stata_dir, stata_flavor = configure_stata_for_tests()
    stata_setup.config(stata_dir, stata_flavor)
except (FileNotFoundError, PermissionError) as e:
    pytest.skip(f"Stata not found or not executable: {e}", allow_module_level=True)


class TestListGraphsTTLCache:
    """Test TTL cache for list_graphs() method."""

    def test_ttl_cache_basic_functionality(self, client, monkeypatch):
        """Test basic TTL cache functionality."""
        monkeypatch.setattr(client, "LIST_GRAPHS_TTL", 1.0, raising=False)
        
        # First call should fetch from Stata
        result1 = client.list_graphs()
        
        # Second call within TTL should use cache (faster)
        start = time.time()
        result2 = client.list_graphs()
        duration = time.time() - start
        
        # Cached call should be very fast (< 100ms)
        assert duration < 0.1
        assert result1 == result2
        
        # Wait for TTL to expire
        time.sleep(1.1)
        
        # Third call after TTL should fetch fresh data
        result3 = client.list_graphs()
        assert result3 == result1  # Same data, but fetched fresh
    
    def test_ttl_cache_invalidation(self, client, monkeypatch):
        """Test cache invalidation functionality."""
        monkeypatch.setattr(client, "LIST_GRAPHS_TTL", 1.0, raising=False)
        
        # First call
        result1 = client.list_graphs()
        
        # Invalidate cache
        client.invalidate_list_graphs_cache()
        
        # Next call should fetch fresh data even within TTL
        result2 = client.list_graphs()
        assert result2 == result1
    
    def test_ttl_cache_error_handling(self, client, monkeypatch):
        """Test TTL cache behavior when Stata calls fail."""
        monkeypatch.setattr(client, "LIST_GRAPHS_TTL", 1.0, raising=False)
        
        # First call succeeds
        result1 = client.list_graphs()
        
        # Cached call should still work
        result2 = client.list_graphs()
        assert result2 == result1
    
    def test_ttl_cache_concurrent_access(self, client, monkeypatch):
        """Test TTL cache under concurrent access."""
        monkeypatch.setattr(client, "LIST_GRAPHS_TTL", 1.0, raising=False)
        
        # Initialize the client fully before starting threads
        # This ensures Stata is ready and prevents race conditions
        _ = client.list_graphs()
        
        results = []
        errors = []
        
        def concurrent_call(thread_id):
            try:
                result = client.list_graphs()
                results.append((thread_id, result))
            except Exception as e:
                errors.append((thread_id, str(e)))
        
        # Start multiple threads
        threads = []
        for thread_id in range(10):
            thread = threading.Thread(target=concurrent_call, args=(thread_id,))
            threads.append(thread)
            thread.start()
        
        # Wait for all threads
        for thread in threads:
            thread.join()
        
        # Verify results
        assert len(errors) == 0
        assert len(results) == 10
        
        # All results should be identical
        expected_result = results[0][1]
        for thread_id, result in results:
            assert result == expected_result
    
    def test_ttl_cache_expiration(self, client, monkeypatch):
        """Test that cache properly expires after TTL."""
        monkeypatch.setattr(client, "LIST_GRAPHS_TTL", 1.0, raising=False)
        
        # First call
        result1 = client.list_graphs()
        
        # Second call within TTL (should be very fast)
        start = time.time()
        result2 = client.list_graphs()
        duration1 = time.time() - start
        assert duration1 < 0.1  # Cached, should be < 100ms
        assert result2 == result1
        
        # Wait for TTL to expire
        time.sleep(0.15)
        
        # Third call after TTL (will be slower as it fetches fresh)
        start = time.time()
        result3 = client.list_graphs()
        duration2 = time.time() - start
        # Fresh fetch should take more time than cache hit
        assert duration2 > duration1
        assert result3 == result1
    
    def test_cache_invalidation_on_graph_creation(self, client, monkeypatch):
        """Test that cache is invalidated when graphs are created."""
        monkeypatch.setattr(client, "LIST_GRAPHS_TTL", 1.0, raising=False)
        
        # Get initial list
        result1 = client.list_graphs()
        initial_count = len(result1)
        
        # Create a simple test graph
        client.run_command_structured("sysuse auto, clear")
        client.run_command_structured("sysuse auto, clear")
        client.run_command_structured("scatter mpg weight")
        
        # Try to cache it (this should invalidate the list cache)
        try:
            client.cache_graph_on_creation("test_graph")
        except:
            pass  # May fail if graph doesn't exist, but cache should be invalidated
        
        # List should now include the new graph (or be refreshed)
        result2 = client.list_graphs()
        # The list might have changed or at least the cache was invalidated
        assert isinstance(result2, list)


if __name__ == "__main__":
    pytest.main([__file__, "-v", "-s"])


"""
Unit tests for dynamic backward log search functionality.

Tests the _read_log_backwards_until_error and _read_log_tail_smart methods.
"""
import os
import tempfile
import pytest
from pathlib import Path

import stata_setup
from conftest import configure_stata_for_tests
try:
    stata_dir, stata_flavor = configure_stata_for_tests()
    stata_setup.config(stata_dir, stata_flavor)
except (FileNotFoundError, PermissionError) as e:
    pytest.skip(f"Stata not found or not executable: {e}", allow_module_level=True)
pytestmark = pytest.mark.requires_stata


class TestBackwardLogSearch:
    """Test the backward search for {err} tags in log files."""

    def test_finds_error_at_end(self, client):
        """Test finding error near the end of the log file."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.log') as f:
            log_path = f.name
            # Write some content
            f.write("Starting analysis...\n" * 100)
            # Error near the end
            f.write("{err}variable compl_gloves not found\n")
            f.write("{txt}cleanup line\n" * 10)

        try:
            result = client._read_log_backwards_until_error(log_path)
            assert '{err}' in result
            assert 'compl_gloves' in result
            assert 'Starting analysis' in result  # Should include earlier context
        finally:
            os.unlink(log_path)

    def test_finds_error_in_middle(self, client):
        """Test finding error in the middle with lots of output after."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.log') as f:
            log_path = f.name
            # Initial content
            f.write("Starting...\n" * 100)
            # Error in middle
            f.write("{err}variable {bf}compl_gloves{sf} not found\n")
            f.write("{txt}      {hline 81} end reghdfe.Estimate {hline}\n")
            # LOTS of cleanup output (simulate your real scenario)
            for i in range(5000):
                f.write(f"    - cleanup operation {i}\n")
            f.write("end of do-file\n")

        try:
            result = client._read_log_backwards_until_error(log_path)
            assert '{err}' in result, "Should find {err} tag"
            assert 'compl_gloves' in result, "Should include error variable name"
        finally:
            os.unlink(log_path)

    def test_finds_error_at_beginning(self, client):
        """Test finding error at the very beginning of the file."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.log') as f:
            log_path = f.name
            # Error right at start
            f.write("{err}syntax error\n")
            f.write("{txt}r(198)\n")
            # Lots of output after
            f.write("cleanup\n" * 10000)

        try:
            result = client._read_log_backwards_until_error(log_path)
            assert '{err}' in result
            assert 'syntax error' in result
        finally:
            os.unlink(log_path)

    def test_multiple_errors_finds_first(self, client):
        """Test that with multiple errors, we find the first one (root cause)."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.log') as f:
            log_path = f.name
            f.write("Starting...\n" * 50)
            # First error (root cause)
            f.write("{err}variable x not found\n")
            f.write("continuing...\n" * 100)
            # Second error (cascading)
            f.write("{err}invalid syntax\n")
            f.write("cleanup\n" * 100)

        try:
            result = client._read_log_backwards_until_error(log_path)
            # Should find at least one error
            assert '{err}' in result
            # Depending on implementation, might find first or last
            # The backward search will find the last error, which is fine
            # A forward search would find the first (root cause)
        finally:
            os.unlink(log_path)

    def test_no_error_returns_all_content(self, client):
        """Test that when no error exists, it returns content up to max_bytes."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.log') as f:
            log_path = f.name
            f.write("No errors here\n" * 100)

        try:
            result = client._read_log_backwards_until_error(log_path, max_bytes=10000)
            assert '{err}' not in result
            assert 'No errors here' in result
        finally:
            os.unlink(log_path)

    def test_empty_file(self, client):
        """Test handling of empty log file."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.log') as f:
            log_path = f.name
            # Write nothing

        try:
            result = client._read_log_backwards_until_error(log_path)
            assert result == ""
        finally:
            os.unlink(log_path)

    def test_respects_max_bytes(self, client):
        """Test that search respects max_bytes limit."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.log') as f:
            log_path = f.name
            # Write 2MB of content without any errors
            for i in range(100000):
                f.write(f"Line {i}: no errors here\n")

        try:
            # Limit to 100KB
            result = client._read_log_backwards_until_error(log_path, max_bytes=100_000)
            assert len(result) <= 110_000  # Some slack for chunk boundaries
        finally:
            os.unlink(log_path)

    def test_handles_unicode(self, client):
        """Test handling of unicode content in logs."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.log', encoding='utf-8') as f:
            log_path = f.name
            f.write("Unicode: café, naïve, 日本語\n" * 50)
            f.write("{err}variable σ not found\n")
            f.write("More unicode: Ω ∑ ∫\n" * 10)

        try:
            result = client._read_log_backwards_until_error(log_path)
            assert '{err}' in result
            assert 'σ' in result
        finally:
            os.unlink(log_path)

    def test_real_stata_error_format(self, client):
        """Test with actual Stata error format from your example."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.log', encoding='utf-8') as f:
            log_path = f.name
            # Simulate the exact format from your error
            f.write(". reghdfe log_total_val i.compl_gloves\n")
            f.write("(MWFE estimator converged in 1 iterations)\n")
            f.write("\n")
            f.write("{err}variable {bf}compl_gloves{sf} not found\n")
            f.write("{txt}      {hline 81} end reghdfe.Estimate {hline}\n")
            f.write("    - Cleanup `c(rc)' `keep_mata'\n")
            f.write("    = Cleanup 111 0\n")
            
            # Add lots of trace output
            for i in range(1000):
                f.write(f"      - trace line {i}\n")
            
            f.write("      - if `rc') exit `rc'\n")
            f.write("      = if (111) exit 111\n")
            f.write("      {hline 82} end reghdfe.Cleanup {hline}\n")
            f.write("r(111);\n")
            f.write("end of do-file\n")
            f.write("r(111);\n")

        try:
            result = client._read_log_backwards_until_error(log_path)
            
            # Critical assertions
            assert '{err}' in result, "Must find {err} tag"
            assert 'compl_gloves' in result, "Must find variable name"
            assert 'variable {bf}compl_gloves{sf} not found' in result, "Must find full error message"
            
            # Should also have some context before the error
            assert 'reghdfe' in result or 'MWFE' in result, "Should have context before error"
            
        finally:
            os.unlink(log_path)


class TestSmartLogReader:
    """Test the smart log reader that adapts based on rc."""

    def test_success_uses_small_tail(self, client):
        """Test that successful runs (rc=0) use small tail."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.log') as f:
            log_path = f.name
            # Write 100KB of content
            for i in range(5000):
                f.write(f"Line {i}: successful output\n")

        try:
            # With rc=0, should only read 20KB tail
            result = client._read_log_tail_smart(log_path, rc=0, trace=False)
            # Result should be much less than 100KB
            assert len(result) < 30_000  # 20KB + some slack
        finally:
            os.unlink(log_path)

    def test_error_uses_backward_search(self, client):
        """Test that errors (rc!=0) use backward search."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.log') as f:
            log_path = f.name
            f.write("Starting...\n" * 100)
            f.write("{err}test error\n")
            # Add 50KB of output after error
            for i in range(2500):
                f.write(f"Cleanup {i}\n")

        try:
            result = client._read_log_tail_smart(log_path, rc=111, trace=False)
            # Should find the error despite lots of output after
            assert '{err}' in result
            assert 'test error' in result
        finally:
            os.unlink(log_path)

    def test_trace_mode_uses_larger_tail(self, client):
        """Test that trace mode uses 200KB tail for successful runs."""
        with tempfile.NamedTemporaryFile(mode='w', delete=False, suffix='.log') as f:
            log_path = f.name
            # Write 300KB of content
            for i in range(15000):
                f.write(f"Line {i}: trace output\n")

        try:
            # With rc=0 and trace=True, should read 200KB tail
            result = client._read_log_tail_smart(log_path, rc=0, trace=True)
            # Should be larger than non-trace but less than full file
            assert 150_000 < len(result) < 250_000
        finally:
            os.unlink(log_path)


class TestErrorExtraction:
    """Test the error extraction from logs with {err} tags."""

    def test_extracts_simple_error(self, client):
        """Test extraction of simple error message."""
        log_content = """
. summarize nonexistent
{err}variable nonexistent not found
{txt}r(111);
"""
        msg, context = client._extract_error_and_context(log_content, rc=111)
        
        assert 'nonexistent' in msg.lower()
        assert '{err}' in context
        assert 'r(111)' in context

    def test_extracts_complex_error_with_smcl(self, client):
        """Test extraction of error with SMCL formatting."""
        log_content = """
. reghdfe log_total_val i.compl_gloves
{err}variable {bf}compl_gloves{sf} not found
{txt}      {hline 81} end reghdfe.Estimate {hline}
r(111);
"""
        msg, context = client._extract_error_and_context(log_content, rc=111)
        
        assert 'compl_gloves' in msg
        assert 'not found' in msg.lower()
        assert '{err}' in context

    def test_extracts_multiline_error(self, client):
        """Test extraction of multi-line error block."""
        log_content = """
. command
{err}error line 1
{err}error line 2  
{err}error line 3
{txt}continuing...
"""
        msg, context = client._extract_error_and_context(log_content, rc=198)
        
        # Should contain all error lines
        assert 'error line 1' in msg or 'error line 1' in context
        assert 'error line 2' in msg or 'error line 2' in context

    def test_fallback_without_err_tag(self, client):
        """Test fallback behavior when no {err} tag present."""
        log_content = """
. command
some output
more output
r(111);
end of do-file
"""
        msg, context = client._extract_error_and_context(log_content, rc=111)
        
        # Should return generic error message
        assert 'Stata error r(111)' in msg
        # Context should contain last lines
        assert 'r(111)' in context


if __name__ == '__main__':
    pytest.main([__file__, '-v'])